/*
 * Copyright (2020) The Delta Lake Project Authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.spark.sql.delta

// scalastyle:off import.ordering.noEmptyLine
import scala.collection.JavaConverters._

import org.apache.spark.sql.delta.actions.Protocol
import org.apache.spark.sql.delta.catalog.{DeltaCatalog, DeltaTableV2}
import org.apache.spark.sql.delta.test.DeltaSQLCommandTest
import org.scalatest.BeforeAndAfter

import org.apache.spark.SparkConf
import org.apache.spark.sql.{AnalysisException, CreateTableWriter, Dataset, QueryTest, Row}
import org.apache.spark.sql.catalyst.analysis.{CannotReplaceMissingTableException, TableAlreadyExistsException}
import org.apache.spark.sql.connector.catalog.{CatalogV2Util, Identifier, Table, TableCatalog}
import org.apache.spark.sql.connector.expressions._
import org.apache.spark.sql.functions._
import org.apache.spark.sql.internal.SQLConf
import org.apache.spark.sql.test.SharedSparkSession
import org.apache.spark.sql.types.{LongType, StringType, StructType}

// These tests are copied from Apache Spark (minus partition by expressions) and should work exactly
// the same with Delta minus some writer options
trait OpenSourceDataFrameWriterV2Tests
    extends QueryTest
    with SharedSparkSession
    with BeforeAndAfter {

  import testImplicits._

  before {
    val df = spark.createDataFrame(Seq((1L, "a"), (2L, "b"), (3L, "c"))).toDF("id", "data")
    df.createOrReplaceTempView("source")
    val df2 = spark.createDataFrame(Seq((4L, "d"), (5L, "e"), (6L, "f"))).toDF("id", "data")
    df2.createOrReplaceTempView("source2")
  }

  after {
    spark.sessionState.catalog.listTables("default").foreach { ti =>
      spark.sessionState.catalog.dropTable(ti, ignoreIfNotExists = false, purge = true)
    }
  }

  def catalog: TableCatalog = {
    spark.sessionState.catalogManager.currentCatalog.asInstanceOf[DeltaCatalog]
  }

  protected def getProperties(table: Table): Map[String, String] = {
    val reservedProp = CatalogV2Util.TABLE_RESERVED_PROPERTIES :+ "Type"
    table.properties().asScala.toMap
      .filterKeys(!reservedProp.contains(_))
      .filterKeys(k =>
        k != Protocol.MIN_READER_VERSION_PROP &&  k != Protocol.MIN_WRITER_VERSION_PROP)
  }

  test("Append: basic append") {
    spark.sql("CREATE TABLE table_name (id bigint, data string) USING delta")

    checkAnswer(spark.table("table_name"), Seq.empty)

    spark.table("source").writeTo("table_name").append()

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(1L, "a"), Row(2L, "b"), Row(3L, "c")))

    spark.table("source2").writeTo("table_name").append()

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(1L, "a"), Row(2L, "b"), Row(3L, "c"), Row(4L, "d"), Row(5L, "e"), Row(6L, "f")))
  }

  test("Append: by name not position") {
    spark.sql("CREATE TABLE table_name (id bigint, data string) USING delta")

    checkAnswer(spark.table("table_name"), Seq.empty)

    val exc = intercept[AnalysisException] {
      spark.table("source").withColumnRenamed("data", "d").writeTo("table_name").append()
    }

    assert(exc.getMessage.contains("schema mismatch"))

    checkAnswer(
      spark.table("table_name"),
      Seq())
  }

  test("Append: fail if table does not exist") {
    val exc = intercept[AnalysisException] {
      spark.table("source").writeTo("table_name").append()
    }

    assert(exc.getMessage.contains("table_name"))
  }

  test("Overwrite: overwrite by expression: true") {
    spark.sql(
      "CREATE TABLE table_name (id bigint, data string) USING delta PARTITIONED BY (id)")

    checkAnswer(spark.table("table_name"), Seq.empty)

    spark.table("source").writeTo("table_name").append()

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(1L, "a"), Row(2L, "b"), Row(3L, "c")))

    spark.table("source2").writeTo("table_name").overwrite(lit(true))

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(4L, "d"), Row(5L, "e"), Row(6L, "f")))
  }

  test("Overwrite: overwrite by expression: id = 3") {
    spark.sql(
      "CREATE TABLE table_name (id bigint, data string) USING delta PARTITIONED BY (id)")

    checkAnswer(spark.table("table_name"), Seq.empty)

    spark.table("source").writeTo("table_name").append()

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(1L, "a"), Row(2L, "b"), Row(3L, "c")))

    val e = intercept[AnalysisException] {
      spark.table("source2").writeTo("table_name").overwrite($"id" === 3)
    }
    assert(e.getMessage.contains("Invalid data would be written to partitions"))

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(1L, "a"), Row(2L, "b"), Row(3L, "c")))
  }

  test("Overwrite: by name not position") {
    spark.sql("CREATE TABLE table_name (id bigint, data string) USING delta")

    checkAnswer(spark.table("table_name"), Seq.empty)

    val exc = intercept[AnalysisException] {
      spark.table("source").withColumnRenamed("data", "d")
        .writeTo("table_name").overwrite(lit(true))
    }

    assert(exc.getMessage.contains("schema mismatch"))

    checkAnswer(
      spark.table("table_name"),
      Seq())
  }

  test("Overwrite: fail if table does not exist") {
    val exc = intercept[AnalysisException] {
      spark.table("source").writeTo("table_name").overwrite(lit(true))
    }

    assert(exc.getMessage.contains("table_name"))
  }

  test("OverwritePartitions: overwrite conflicting partitions") {
    spark.sql(
      "CREATE TABLE table_name (id bigint, data string) USING delta PARTITIONED BY (id)")

    checkAnswer(spark.table("table_name"), Seq.empty)

    spark.table("source").writeTo("table_name").append()

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(1L, "a"), Row(2L, "b"), Row(3L, "c")))

    val e = intercept[AnalysisException] {
      spark.table("source2").withColumn("id", $"id" - 2)
        .writeTo("table_name").overwritePartitions()
    }
    assert(e.getMessage.contains("Table default.table_name does not support dynamic overwrite"))

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(1L, "a"), Row(2L, "b"), Row(3L, "c")))
  }

  test("OverwritePartitions: overwrite all rows if not partitioned") {
    spark.sql("CREATE TABLE table_name (id bigint, data string) USING delta")

    checkAnswer(spark.table("table_name"), Seq.empty)

    spark.table("source").writeTo("table_name").append()

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(1L, "a"), Row(2L, "b"), Row(3L, "c")))

    val e = intercept[AnalysisException] {
      spark.table("source2").writeTo("table_name").overwritePartitions()
    }
    assert(e.getMessage.contains("Table default.table_name does not support dynamic overwrite"))
  }

  test("OverwritePartitions: by name not position") {
    spark.sql("CREATE TABLE table_name (id bigint, data string) USING delta")

    checkAnswer(spark.table("table_name"), Seq.empty)

    val e = intercept[AnalysisException] {
      spark.table("source").withColumnRenamed("data", "d")
        .writeTo("table_name").overwritePartitions()
    }

    assert(e.getMessage.contains("Table default.table_name does not support dynamic overwrite"))

    checkAnswer(
      spark.table("table_name"),
      Seq())
  }

  test("OverwritePartitions: fail if table does not exist") {
    val exc = intercept[AnalysisException] {
      spark.table("source").writeTo("table_name").overwritePartitions()
    }

    assert(exc.getMessage.contains("table_name"))
  }

  test("Create: basic behavior") {
    spark.table("source").writeTo("table_name").using("delta").create()

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(1L, "a"), Row(2L, "b"), Row(3L, "c")))

    val table = catalog.loadTable(Identifier.of(Array("default"), "table_name"))

    assert(table.name === "default.table_name")
    assert(table.schema === new StructType().add("id", LongType).add("data", StringType))
    assert(table.partitioning.isEmpty)
    assert(getProperties(table).isEmpty)
  }

  test("Create: with using") {
    spark.table("source").writeTo("table_name").using("delta").create()

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(1L, "a"), Row(2L, "b"), Row(3L, "c")))

    val table = catalog.loadTable(Identifier.of(Array("default"), "table_name"))

    assert(table.name === "default.table_name")
    assert(table.schema === new StructType().add("id", LongType).add("data", StringType))
    assert(table.partitioning.isEmpty)
    assert(getProperties(table).isEmpty)
  }

  test("Create: with property") {
    spark.table("source").writeTo("table_name")
      .tableProperty("prop", "value").using("delta").create()

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(1L, "a"), Row(2L, "b"), Row(3L, "c")))

    val table = catalog.loadTable(Identifier.of(Array("default"), "table_name"))

    assert(table.name === "default.table_name")
    assert(table.schema === new StructType().add("id", LongType).add("data", StringType))
    assert(table.partitioning.isEmpty)
    assert(getProperties(table) === Map("prop" -> "value"))
  }

  test("Create: identity partitioned table") {
    spark.table("source").writeTo("table_name").using("delta").partitionedBy($"id").create()

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(1L, "a"), Row(2L, "b"), Row(3L, "c")))

    val table = catalog.loadTable(Identifier.of(Array("default"), "table_name"))

    assert(table.name === "default.table_name")
    assert(table.schema === new StructType().add("id", LongType).add("data", StringType))
    assert(table.partitioning === Seq(IdentityTransform(FieldReference("id"))))
    assert(getProperties(table).isEmpty)
  }

  test("Create: fail if table already exists") {
    spark.sql(
      "CREATE TABLE table_name (id bigint, data string) USING delta PARTITIONED BY (id)")

    val exc = intercept[TableAlreadyExistsException] {
      spark.table("source").writeTo("table_name").using("delta").create()
    }

    assert(exc.getMessage.contains("table_name"))

    val table = catalog.loadTable(Identifier.of(Array("default"), "table_name"))

    // table should not have been changed
    assert(table.name === "default.table_name")
    assert(table.schema === new StructType().add("id", LongType).add("data", StringType))
    assert(table.partitioning === Seq(IdentityTransform(FieldReference("id"))))
    assert(getProperties(table).isEmpty)
  }

  test("Replace: basic behavior") {
    spark.sql(
      "CREATE TABLE table_name (id bigint, data string) USING delta PARTITIONED BY (id)")
    spark.sql("INSERT INTO TABLE table_name SELECT * FROM source")

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(1L, "a"), Row(2L, "b"), Row(3L, "c")))

    val table = catalog.loadTable(Identifier.of(Array("default"), "table_name"))

    // validate the initial table
    assert(table.name === "default.table_name")
    assert(table.schema === new StructType().add("id", LongType).add("data", StringType))
    assert(table.partitioning === Seq(IdentityTransform(FieldReference("id"))))
    assert(getProperties(table).isEmpty)

    spark.table("source2")
      .withColumn("even_or_odd", when(($"id" % 2) === 0, "even").otherwise("odd"))
      .writeTo("table_name").using("delta")
      .tableProperty("deLta.aPpeNdonly", "true").replace()

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(4L, "d", "even"), Row(5L, "e", "odd"), Row(6L, "f", "even")))

    val replaced = catalog.loadTable(Identifier.of(Array("default"), "table_name"))

    // validate the replacement table
    assert(replaced.name === "default.table_name")
    assert(replaced.schema === new StructType()
      .add("id", LongType)
      .add("data", StringType)
      .add("even_or_odd", StringType))
    assert(replaced.partitioning.isEmpty)
    assert(getProperties(replaced) === Map("delta.appendOnly" -> "true"))
  }

  test("Replace: partitioned table") {
    spark.sql("CREATE TABLE table_name (id bigint, data string) USING delta")
    spark.sql("INSERT INTO TABLE table_name SELECT * FROM source")

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(1L, "a"), Row(2L, "b"), Row(3L, "c")))

    val table = catalog.loadTable(Identifier.of(Array("default"), "table_name"))

    // validate the initial table
    assert(table.name === "default.table_name")
    assert(table.schema === new StructType().add("id", LongType).add("data", StringType))
    assert(table.partitioning.isEmpty)
    assert(getProperties(table).isEmpty)

    spark.table("source2")
      .withColumn("even_or_odd", when(($"id" % 2) === 0, "even").otherwise("odd"))
      .writeTo("table_name").using("delta")
      .partitionedBy($"id")
      .replace()

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(4L, "d", "even"), Row(5L, "e", "odd"), Row(6L, "f", "even")))

    val replaced = catalog.loadTable(Identifier.of(Array("default"), "table_name"))

    // validate the replacement table
    assert(replaced.name === "default.table_name")
    assert(replaced.schema === new StructType()
      .add("id", LongType)
      .add("data", StringType)
      .add("even_or_odd", StringType))
    assert(replaced.partitioning === Seq(IdentityTransform(FieldReference("id"))))
    assert(getProperties(replaced).isEmpty)
  }

  test("Replace: fail if table does not exist") {
    val exc = intercept[CannotReplaceMissingTableException] {
      spark.table("source").writeTo("table_name").using("delta").replace()
    }

    assert(exc.getMessage.contains("table_name"))
  }

  test("CreateOrReplace: table does not exist") {
    spark.table("source2").writeTo("table_name").using("delta").createOrReplace()

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(4L, "d"), Row(5L, "e"), Row(6L, "f")))

    val replaced = catalog.loadTable(Identifier.of(Array("default"), "table_name"))

    // validate the replacement table
    assert(replaced.name === "default.table_name")
    assert(replaced.schema === new StructType().add("id", LongType).add("data", StringType))
    assert(replaced.partitioning.isEmpty)
    assert(getProperties(replaced).isEmpty)
  }

  test("CreateOrReplace: table exists") {
    spark.sql(
      "CREATE TABLE table_name (id bigint, data string) USING delta PARTITIONED BY (id)")
    spark.sql("INSERT INTO TABLE table_name SELECT * FROM source")

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(1L, "a"), Row(2L, "b"), Row(3L, "c")))

    val table = catalog.loadTable(Identifier.of(Array("default"), "table_name"))

    // validate the initial table
    assert(table.name === "default.table_name")
    assert(table.schema === new StructType().add("id", LongType).add("data", StringType))
    assert(table.partitioning === Seq(IdentityTransform(FieldReference("id"))))
    assert(getProperties(table).isEmpty)

    spark.table("source2")
      .withColumn("even_or_odd", when(($"id" % 2) === 0, "even").otherwise("odd"))
      .writeTo("table_name").using("delta").createOrReplace()

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(4L, "d", "even"), Row(5L, "e", "odd"), Row(6L, "f", "even")))

    val replaced = catalog.loadTable(Identifier.of(Array("default"), "table_name"))

    // validate the replacement table
    assert(replaced.name === "default.table_name")
    assert(replaced.schema === new StructType()
      .add("id", LongType)
      .add("data", StringType)
      .add("even_or_odd", StringType))
    assert(replaced.partitioning.isEmpty)
    assert(getProperties(replaced).isEmpty)
  }

  test("Create: partitioned by years(ts) - not supported") {
    val e = intercept[AnalysisException] {
      spark.table("source")
        .withColumn("ts", lit("2019-06-01 10:00:00.000000").cast("timestamp"))
        .writeTo("table_name")
        .partitionedBy(years($"ts"))
        .using("delta")
        .create()
    }
    assert(e.getMessage.contains("Partitioning by expressions"))
  }

  test("Create: partitioned by months(ts) - not supported") {
    val e = intercept[AnalysisException] {
      spark.table("source")
        .withColumn("ts", lit("2019-06-01 10:00:00.000000").cast("timestamp"))
        .writeTo("table_name")
        .partitionedBy(months($"ts"))
        .using("delta")
        .create()
    }
    assert(e.getMessage.contains("Partitioning by expressions"))
  }

  test("Create: partitioned by days(ts) - not supported") {
    val e = intercept[AnalysisException] {
      spark.table("source")
        .withColumn("ts", lit("2019-06-01 10:00:00.000000").cast("timestamp"))
        .writeTo("table_name")
        .partitionedBy(days($"ts"))
        .using("delta")
        .create()
    }
    assert(e.getMessage.contains("Partitioning by expressions"))
  }

  test("Create: partitioned by hours(ts) - not supported") {
    val e = intercept[AnalysisException] {
      spark.table("source")
        .withColumn("ts", lit("2019-06-01 10:00:00.000000").cast("timestamp"))
        .writeTo("table_name")
        .partitionedBy(hours($"ts"))
        .using("delta")
        .create()
    }
    assert(e.getMessage.contains("Partitioning by expressions"))
  }

  test("Create: partitioned by bucket(4, id) - not supported") {
    val e = intercept[AnalysisException] {
      spark.table("source")
        .writeTo("table_name")
        .partitionedBy(bucket(4, $"id"))
        .using("delta")
        .create()
    }
    assert(e.getMessage.contains("Bucketing"))
  }
}

class DeltaDataFrameWriterV2Suite
  extends OpenSourceDataFrameWriterV2Tests
  with DeltaSQLCommandTest {

  import testImplicits._

  test("Append: basic append by path") {
    spark.sql("CREATE TABLE table_name (id bigint, data string) USING delta")

    checkAnswer(spark.table("table_name"), Seq.empty)
    val location = catalog.loadTable(Identifier.of(Array("default"), "table_name"))
      .asInstanceOf[DeltaTableV2].path

    spark.table("source").writeTo(s"delta.`$location`").append()

    checkAnswer(
      spark.table(s"delta.`$location`"),
      Seq(Row(1L, "a"), Row(2L, "b"), Row(3L, "c")))
  }

  test("Create: basic behavior by path") {
    withTempDir { tempDir =>
      val dir = tempDir.getCanonicalPath
      spark.table("source").writeTo(s"delta.`$dir`").using("delta").create()

      checkAnswer(
        spark.read.format("delta").load(dir),
        Seq(Row(1L, "a"), Row(2L, "b"), Row(3L, "c")))

      val table = catalog.loadTable(Identifier.of(Array("delta"), dir))

      assert(table.name === s"delta.`file:$dir`")
      assert(table.schema === new StructType().add("id", LongType).add("data", StringType))
      assert(table.partitioning.isEmpty)
      assert(getProperties(table).isEmpty)
    }
  }

  test("Create: using empty dataframe") {
    spark.table("source").where("false")
      .writeTo("table_name").using("delta")
      .tableProperty("delta.appendOnly", "true")
      .partitionedBy($"id").create()

    checkAnswer(spark.table("table_name"), Seq.empty[Row])

    val table = catalog.loadTable(Identifier.of(Array("default"), "table_name"))

    assert(table.name === "default.table_name")
    assert(table.schema === new StructType().add("id", LongType).add("data", StringType))
    assert(table.partitioning === Seq(IdentityTransform(FieldReference("id"))))
    assert(getProperties(table) === Map("delta.appendOnly" -> "true"))
  }

  test("Replace: basic behavior using empty df") {
    spark.sql(
      "CREATE TABLE table_name (id bigint, data string) USING delta PARTITIONED BY (id)")
    spark.sql("INSERT INTO TABLE table_name SELECT * FROM source")

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(1L, "a"), Row(2L, "b"), Row(3L, "c")))

    val table = catalog.loadTable(Identifier.of(Array("default"), "table_name"))

    // validate the initial table
    assert(table.name === "default.table_name")
    assert(table.schema === new StructType().add("id", LongType).add("data", StringType))
    assert(table.partitioning === Seq(IdentityTransform(FieldReference("id"))))
    assert(getProperties(table).isEmpty)

    spark.table("source2").where("false")
      .withColumn("even_or_odd", when(($"id" % 2) === 0, "even").otherwise("odd"))
      .writeTo("table_name").using("delta")
      .tableProperty("deLta.aPpeNdonly", "true").replace()

    checkAnswer(
      spark.table("table_name"),
      Seq.empty[Row])

    val replaced = catalog.loadTable(Identifier.of(Array("default"), "table_name"))

    // validate the replacement table
    assert(replaced.name === "default.table_name")
    assert(replaced.schema === new StructType()
        .add("id", LongType)
        .add("data", StringType)
        .add("even_or_odd", StringType))
    assert(replaced.partitioning.isEmpty)
    assert(getProperties(replaced) === Map("delta.appendOnly" -> "true"))
  }

  test("throw error with createOrReplace and Replace if overwriteSchema=false") {
    spark.sql(
      "CREATE TABLE table_name (id bigint, data string) USING delta PARTITIONED BY (id)")
    spark.sql("INSERT INTO TABLE table_name SELECT * FROM source")

    checkAnswer(
      spark.table("table_name"),
      Seq(Row(1L, "a"), Row(2L, "b"), Row(3L, "c")))

    def checkFailure(
        df: Dataset[_],
        errorMsg: String)(
        f: CreateTableWriter[_] => CreateTableWriter[_]): Unit = {
      val e = intercept[IllegalArgumentException] {
        val dfwV2 = df.writeTo("table_name")
          .using("delta")
          .option("overwriteSchema", "false")
        f(dfwV2).replace()
      }
      assert(e.getMessage.contains(errorMsg))

      val e2 = intercept[IllegalArgumentException] {
        val dfwV2 = df.writeTo("table_name")
            .using("delta")
            .option("overwriteSchema", "false")
        f(dfwV2).createOrReplace()
      }
      assert(e2.getMessage.contains(errorMsg))
    }

    // schema changes
    checkFailure(
      spark.table("table_name").withColumn("id2", 'id + 1),
      "overwriteSchema is not allowed when replacing")(a => a.partitionedBy($"id"))

    // partitioning changes
    // did not specify partitioning
    checkFailure(spark.table("table_name"),
      "overwriteSchema is not allowed when replacing")(a => a)

    // different partitioning column
    checkFailure(spark.table("table_name"),
      "overwriteSchema is not allowed when replacing")(a => a.partitionedBy($"data"))

    // different table Properties
    checkFailure(spark.table("table_name"), "overwriteSchema is not allowed when replacing")(a =>
      a.partitionedBy($"id").tableProperty("delta.appendOnly", "true"))
  }
}
